# https://github.com/apache/airflow/issues/7870

try:
  from pydantic import create_model
  import uvicorn

  from fastapi import FastAPI
  from fastapi.responses import JSONResponse, Response
  from fastapi.middleware.cors import CORSMiddleware
except ImportError:
  # if this is happening to you sir, why don't you come work with us?
  FastAPI = None

import json
import inspect
from typing import Any, Dict

from nbox.version import __version__
from nbox.operator import Operator
from nbox.nbxlib.operator_spec import OperatorType
from nbox.utils import py_from_bs64, py_to_bs64, logger

def serve_operator(
  op: Operator,
  host: str = "0.0.0.0",
  port: int = 8000,
  *,
  log_system_metrics: bool = True,
  log_user_io: bool = False,
  model_name: str = ""
):
  if FastAPI is None:
    logger.error("To run servers you will need to install the relevant dependencies:")
    logger.error("  pip install -U nbox[serving]")
    raise ImportError("fastapi not installed")

  app = FastAPI()
  app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
  )

  # define ping endpoint
  async def ping_fn():
    return {"message": "pong"}
  app.add_api_route("/", ping_fn, methods=["GET"], response_class=JSONResponse)

  # define metadata endpoint
  async def metadata():
    return {"metadata": {"name": model_name}}
  app.add_api_route("/metadata", metadata, methods=["GET"], response_class=JSONResponse)

  # a special route for Operators to communicate with each other
  async def who_are_you():
    return {"name": op.__qualname__, "nbox_version": __version__}
  app.add_api_route("/who_are_you", who_are_you, methods=["GET"], response_class=JSONResponse)

  for route, fn in get_fastapi_routes(op):
    app.add_api_route(route, fn, methods=["POST"], response_class=JSONResponse)

  # if log_system_metrics:
  #   app.add_middleware(LmaoASGIMiddleware)

  uvicorn.run(app, host = host, port = port)

def get_fastapi_routes(op: Operator):
  """To keep seperation of responsibility the paths are scoped out like all the functions are
  in the /method_{...} and all the custom python code is in /nbx_py_rpc"""
  if op._op_type == OperatorType.WRAP_CLS:
    routes = []
    # add functions that the user has exposed
    wrap_class = op._op_spec.wrap_obj
    for p in dir(wrap_class.__class__):
      if p.startswith("__"):
        continue
      fn = getattr(wrap_class, p)
      routes.append((f"/method_{p}", get_fastapi_fn(fn)))
      routes.append((f"/method_{p}_rest", get_fastapi_fn(fn, _rest = True)))

    # add functions that the python itself can support
    routes.append((f"/nbx_py_rpc", nbx_py_rpc(op)))
  elif op._op_type in [OperatorType.JOB, OperatorType.SERVING]:
    raise RuntimeError("Cannot serve a job or serving operator")
  else:
    routes = [
      ("/forward", get_fastapi_fn(op.forward)),
      ("/forward_rest", get_fastapi_fn(op.forward, _rest = True)),
    ]
  return routes


# builder method is used to progrmatically generate api routes related information for the fastapi app
def get_fastapi_fn(fn, _rest = False):
  from pydantic import create_model
  
  # we use inspect signature instead of writing our own ast thing
  signature = inspect.signature(fn)
  data_dict = {}
  for param in signature.parameters.values():
    default = param.default
    annot = param.annotation
    if default == inspect._empty:
      default = None
    if param.annotation == inspect._empty:
      annot = Any
    data_dict[param.name] = (annot, default)

  # if your function takes in inputs then it is expected to be sent as query params, so create a pydantic
  # model and FastAPI will take care of the rest
  name = f"{fn.__name__}_Request"
  if _rest:
    name = f"{fn.__name__}_Rest_Request"
  base_model = create_model(name, **data_dict)
  base_model_rpc = create_model(name, **{k:(str, py_to_bs64(v[1])) for k,v in data_dict.items()})

  # pretty simple forward function, note that it gets operator using get_op which will be a cache hit
  async def generic_fwd(req: base_model_rpc, response: Response):
    # need to add serialisation to this function because user won't by default send in a serialised object
    data = req.dict()
    try:
      data = {k: py_from_bs64(v) for k, v in data.items()}
    except Exception as e:
      logger.error(f"Failed to convert {data} to python object")
      logger.error(e)
      response.status_code = 400
      return {"error": str(e)}

    try:
      out = fn(**data)
      return {"success": True, "value": py_to_bs64(out)}
    except Exception as e:
      response.status_code = 500
      return {"success": False, "message": str(e)}

  # pretty simple forward function for REST endpoints
  async def generic_fwd_rest(req: base_model, response: Response):
    # need to add serialisation to this function because user won't by default send in a serialised object
    data = req.dict()
    try:
      out = fn(**data)
      try:
        _ = json.dumps(out)
      except:
        response.status_code = 500
        return {"success": False, "message": "Function output cannot be serialised to JSON"}
      return {"success": True, "value": out}
    except Exception as e:
      response.status_code = 500
      return {"success": False, "message": str(e)}

  return generic_fwd_rest if _rest else generic_fwd


def nbx_py_rpc(op: Operator):
  base_model = create_model("nbx_py_rpc", rpc_name = (str, ""), key = (str, ""), value = (str, ""),)
  _nbx_py_rpc = NbxPyRpc(op)
  
  async def forward(req: base_model, response: Response):
    # no need to add serialisation because the NbPyRpc class will handle it
    data = req.dict()
    return _nbx_py_rpc(data, response)

  return forward


class NbxPyRpc(Operator):
  """This object is a shallow class that is used as a router of functions. Distributed computing combined with
  user friendliness of python means that some methods acn be routed and managed as long as there is a wire
  protocol. So here it is:

  request = {
    "rpc_name": "__getattr__",
    "key": "string",            # always there
    "value": "b64-cloudpickle", # optional for functions
  }

  response = {
    "success": bool,
    "message": str,   # optional, will be error in case of failure
    "value": str,     # optional, will be b64-cloudpickle or might be empty ex. del
  }

  we should also add some routes to support a subset of important language features:

  1. __getattr__: obtain any value by doing: `obj.x`
  2. __getitem__: obtain any value by doing: `obj[x]`
  3. __setitem__: set any value by doing: `obj[x] = y`
  4. __delitem__: delete any value by doing: `del obj[x]`
  5. __iter__: iterate over any iterable by doing: `for x in obj`
  6. __next__: get next value from an iterator by doing: `next(obj)`
  7. __len__: get length of any object by doing: `len(obj)`
  8. __contains__: check if an object contains a value by doing: `x in obj`

  The reason we have chosen these for starting is that they can be used to represent any
  data structure required and get/set information from it. We can add more later like
  __enter__ and __exit__ to support context managers. Others like numerical operations
  __add__ and __sub__ doesn't really make sense. Maybe one day when we have neural networks
  but even then it's not clear how we would use them.
  """
  def __init__(self, op: Operator):
    super().__init__()
    self.wrapped_cls = op

  def forward(self, data, response: Response) -> Dict[str, str]:
    _k = set(tuple(data.keys())) - set(["rpc_name", "key", "value"])
    if _k:
      response.status_code = 400
      return {"success": False, "message": f"invalid keys: {_k}"}
    rpc_name = data.get("rpc_name", "")
    fn_map = {
      "__getattr__": (self.fn_getattr, key),
      "__getitem__": (self.fn_getitem, key),
      "__setitem__": (self.fn_setitem, key, value),
      "__delitem__": (self.fn_delitem, key),
      "__iter__": (self.fn_iter),
      "__next__": (self.fn_next),
      "__len__": (self.fn_len),
      "__contains__": (self.fn_contains, key),
    }
    _items = fn_map.get(rpc_name, None)
    if _items is None:
      response.status_code = 400
      return {"success": False, "message": f"invalid rpc_name: {rpc_name}"}

    key = data.get("key", "")
    value = data.get("value", "")

    if key:
      key = py_from_bs64(key)
    if value:
      value = py_from_bs64(value)

    fn, *args = _items
    try:
      out = fn(*args)
      return out
    except Exception as e:
      response.status_code = 500
      return {"success": False, "message": str(e)}
  
  def fn_getattr(self, key):
    out = getattr(self.wrapped_cls._op_wrap, key)
    return {"success": True, "value": py_to_bs64(out)}

  def fn_getitem(self, key):
    out = self.wrapped_cls._op_wrap[key]
    return {"success": True, "value": py_to_bs64(out)}
  
  def fn_setitem(self, key, value):
    self.wrapped_cls._op_wrap[key] = value
    return {"success": True}

  def fn_delitem(self, key):
    del self.wrapped_cls._op_wrap[key]
    return {"success": True}

  def fn_iter(self):
    out = iter(self.wrapped_cls._op_wrap)
    return {"success": True, "value": py_to_bs64(out)}

  def fn_next(self):
    out = next(self.wrapped_cls._op_wrap)
    return {"success": True, "value": py_to_bs64(out)}

  def fn_len(self):
    out = len(self.wrapped_cls._op_wrap)
    return {"success": True, "value": py_to_bs64(out)}

  def fn_contains(self, key):
    out = key in self.wrapped_cls._op_wrap
    return {"success": True, "value": py_to_bs64(out)}



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import job_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Serving(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _DeploymentType:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _DeploymentTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Serving._DeploymentType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNKNOWN: Serving._DeploymentType.ValueType  # 0
        """Unknown deployment type"""

        NBOX_SERVING: Serving._DeploymentType.ValueType  # 1
        """vanilla nbox serving with fastapi"""

    class DeploymentType(_DeploymentType, metaclass=_DeploymentTypeEnumTypeWrapper):
        """Deployment type"""
        pass

    UNKNOWN: Serving.DeploymentType.ValueType  # 0
    """Unknown deployment type"""

    NBOX_SERVING: Serving.DeploymentType.ValueType  # 1
    """vanilla nbox serving with fastapi"""


    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    AUTH_INFO_FIELD_NUMBER: builtins.int
    CREATED_AT_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    RESOURCE_FIELD_NUMBER: builtins.int
    id: typing.Text
    """ID for the deployment"""

    name: typing.Text
    """name of the deployment"""

    @property
    def auth_info(self) -> job_pb2.NBXAuthInfo:
        """string workspace_id = 3; // Workspace ID, if None assume in user workspace
        auth info for the deployment
        """
        pass
    @property
    def created_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """time created"""
        pass
    type: global___Serving.DeploymentType.ValueType
    @property
    def resource(self) -> job_pb2.Resource:
        """deployment resource"""
        pass
    def __init__(self,
        *,
        id: typing.Text = ...,
        name: typing.Text = ...,
        auth_info: typing.Optional[job_pb2.NBXAuthInfo] = ...,
        created_at: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        type: global___Serving.DeploymentType.ValueType = ...,
        resource: typing.Optional[job_pb2.Resource] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","created_at",b"created_at","resource",b"resource"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","created_at",b"created_at","id",b"id","name",b"name","resource",b"resource","type",b"type"]) -> None: ...
global___Serving = Serving



"""
All the Operators that NimbleBox uses are defined in this module.
"""

import os
import zipfile
from datetime import datetime, timezone
from typing import Iterable

from nbox import Operator, logger
from nbox.instance import Instance


class FilesToInstance(Operator):
  def __init__(self, i: str, folder_name: str, workspace_id: str = None) -> None:
    """Transfer any number of files to a NimbleBox Instance at the `folder_name`.

    Args:
      i (str): Instance ID or Name
      folder_name (str): folder to transfer the zip object to
      workspace_id (str): Workspace ID or name, set to `None` for personal workspace
    """
    super().__init__()
    self.folder_name = folder_name
    logger.warn(f"Please ensure that the folder {self.folder_name} exists in '{i}'")
    self.instance = Instance(i, workspace_id)

  def forward(self, files: Iterable[str]):
    """Args:
        files (Iterable[str]): List of all the files to be transfered
    """
    # if not running, start the instance with the basic configuration, we just need to move the files
    self.instance.start()

    # pack all the items in a single zip file
    zip_name = datetime.now(timezone.utc).strftime("%Y-%m-%d-%H-%M") + ".zip"
    logger.info(f"Packing to path: {zip_name}")
    with zipfile.ZipFile(zip_name, 'w', zipfile.ZIP_DEFLATED) as zip_file:
      for f in files:
        zip_file.write(filename = f, arcname = f)

    # sanity check for the zip file
    os_stat = os.stat(zip_name)
    logger.info(f"Zip file size: {os_stat.st_size}")

    self.instance.mv(zip_name, f"nbx://{self.folder_name}/{zip_name}")



"""This file contains the code for observability in the networking layer. This is not the same as in case of prometheus
because in our structure the pod is going to be informing the DB about the metrics and not the other way around. We are
chosing this approach because it allows for more flexibility in the future."""

from typing import Tuple
from threading import Lock
import time

try:
  import starlette
  from starlette.requests import Request
  from starlette.responses import Response
  from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
  from starlette.routing import Match
  from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR
  from starlette.types import ASGIApp
except ImportError:
  starlette = None


class GenericMetric:
  # this is like the prometheus client's Counter, Gauge, Histogram, etc. but instead of client
  # worrying about the types, here we are saying that it's the servers headache to worry about
  def __init__(self, name: str, description: str = ""):
    self.name = name
    self.description = description
    self._value = 0
    self._mutex = Lock()

  def inc(self, value: int = 1):
    with self._mutex:
      self._value += value


class AsgiLmaoMiddleware(BaseHTTPMiddleware):
  def __init__(self, app: ASGIApp, filter_unhandled_paths: bool = False) -> None:
    if starlette is None:
      raise ValueError("Starlette is not installed. pip install -U nbox\[serving\]")
    super().__init__(app)
    self.filter_unhandled_paths = filter_unhandled_paths

  async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
    method = request.method
    path_template, is_handled_path = self.get_path_template(request)

    if self._is_path_filtered(is_handled_path):
      return await call_next(request)

    REQUESTS_IN_PROGRESS.labels(method = method, path_template = path_template).inc()
    REQUESTS.labels(method=method, path_template=path_template).inc()
    before_time = time.perf_counter()
    try:
      response = await call_next(request)
    except BaseException as e:
      status_code = HTTP_500_INTERNAL_SERVER_ERROR
      EXCEPTIONS.labels(method=method, path_template=path_template, exception_type=type(e).__name__).inc()
      raise e from None
    else:
      status_code = response.status_code
      after_time = time.perf_counter()
      REQUESTS_PROCESSING_TIME.labels(method=method, path_template=path_template).observe(after_time - before_time)
    finally:
      RESPONSES.labels(method=method, path_template=path_template, status_code=status_code).inc()
      REQUESTS_IN_PROGRESS.labels(method=method, path_template=path_template).dec()

    return response

  @staticmethod
  def get_path_template(request: Request) -> Tuple[str, bool]:
    for route in request.app.routes:
      match, child_scope = route.matches(request.scope)
      if match == Match.FULL:
        return route.path, True
    return request.url.path, False

  def _is_path_filtered(self, is_handled_path: bool) -> bool:
    return self.filter_unhandled_paths and not is_handled_path





# def metrics(request: Request) -> Response:
#   if "prometheus_multiproc_dir" in os.environ:
#     registry = CollectorRegistry()
#     MultiProcessCollector(registry)
#   else:
#     registry = REGISTRY

#   return Response(generate_latest(registry), headers={"Content-Type": CONTENT_TYPE_LATEST})



from starlette_prometheus import metrics, PrometheusMiddleware



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nbox_ws.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import job_pb2 as job__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from google.protobuf import field_mask_pb2 as google_dot_protobuf_dot_field__mask__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rnbox_ws.proto\x12\x07jobs.ws\x1a\tjob.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1egoogle/protobuf/wrappers.proto\x1a google/protobuf/field_mask.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"1\n\x07JobInfo\x12\x0e\n\x02id\x18\x01 \x01(\tB\x02\x18\x01\x12\x16\n\x03job\x18\x02 \x01(\x0b\x32\t.jobs.Job\"^\n\x0fListJobsRequest\x12$\n\tauth_info\x18\x01 \x01(\x0b\x32\x11.jobs.NBXAuthInfo\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\"D\n\x10ListJobsResponse\x12\x17\n\x04Jobs\x18\x01 \x03(\x0b\x32\t.jobs.Job\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"L\n\x11UploadCodeRequest\x12\x16\n\x03job\x18\x01 \x01(\x0b\x32\t.jobs.Job\x12\x1f\n\x04\x61uth\x18\x02 \x01(\x0b\x32\x11.jobs.NBXAuthInfo\"*\n\x10\x43reateJobRequest\x12\x16\n\x03job\x18\x03 \x01(\x0b\x32\t.jobs.Job\"[\n\x10UpdateJobRequest\x12\x16\n\x03job\x18\x01 \x01(\x0b\x32\t.jobs.Job\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"_\n\x0eJobLogsRequest\x12\x1d\n\x03job\x18\x01 \x01(\x0b\x32\x10.jobs.ws.JobInfo\x12\x1f\n\x08k8s_info\x18\x02 \x01(\x0b\x32\r.jobs.K8sInfo\x12\r\n\x05lines\x18\x03 \x01(\x05\"U\n\x06JobLog\x12\x1d\n\x03job\x18\x01 \x01(\x0b\x32\x10.jobs.ws.JobInfo\x12\x1f\n\x08k8s_info\x18\x02 \x01(\x0b\x32\r.jobs.K8sInfo\x12\x0b\n\x03log\x18\x03 \x03(\t\"z\n\x10UpdateRunRequest\x12\r\n\x05token\x18\x01 \x01(\t\x12\x16\n\x03job\x18\x02 \x01(\x0b\x32\t.jobs.Job\x12.\n\nupdated_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07message\x18\x04 \x01(\t2\xd4\x04\n\x0cWSJobService\x12\'\n\x06GetJob\x12\x10.jobs.ws.JobInfo\x1a\t.jobs.Job\"\x00\x12\x41\n\x08ListJobs\x12\x18.jobs.ws.ListJobsRequest\x1a\x19.jobs.ws.ListJobsResponse\"\x00\x12\x38\n\rUploadJobCode\x12\x1a.jobs.ws.UploadCodeRequest\x1a\t.jobs.Job\"\x00\x12\x33\n\tCreateJob\x12\x19.jobs.ws.CreateJobRequest\x1a\t.jobs.Job\"\x00\x12\x33\n\tUpdateJob\x12\x19.jobs.ws.UpdateJobRequest\x1a\t.jobs.Job\"\x00\x12\x37\n\tDeleteJob\x12\x10.jobs.ws.JobInfo\x1a\x16.google.protobuf.Empty\"\x00\x12:\n\nGetJobLogs\x12\x17.jobs.ws.JobLogsRequest\x1a\x0f.jobs.ws.JobLog\"\x00\x30\x01\x12<\n\nTriggerJob\x12\x10.jobs.ws.JobInfo\x1a\x1a.google.protobuf.BoolValue\"\x00\x12@\n\tUpdateRun\x12\x19.jobs.ws.UpdateRunRequest\x1a\x16.google.protobuf.Empty\"\x00\x12?\n\x0bGetRunToken\x12\x10.jobs.ws.JobInfo\x1a\x1c.google.protobuf.StringValue\"\x00\x42\x32Z0github.com/NimbleBoxAI/jobs-architecture/jobs_pbb\x06proto3')



_JOBINFO = DESCRIPTOR.message_types_by_name['JobInfo']
_LISTJOBSREQUEST = DESCRIPTOR.message_types_by_name['ListJobsRequest']
_LISTJOBSRESPONSE = DESCRIPTOR.message_types_by_name['ListJobsResponse']
_UPLOADCODEREQUEST = DESCRIPTOR.message_types_by_name['UploadCodeRequest']
_CREATEJOBREQUEST = DESCRIPTOR.message_types_by_name['CreateJobRequest']
_UPDATEJOBREQUEST = DESCRIPTOR.message_types_by_name['UpdateJobRequest']
_JOBLOGSREQUEST = DESCRIPTOR.message_types_by_name['JobLogsRequest']
_JOBLOG = DESCRIPTOR.message_types_by_name['JobLog']
_UPDATERUNREQUEST = DESCRIPTOR.message_types_by_name['UpdateRunRequest']
JobInfo = _reflection.GeneratedProtocolMessageType('JobInfo', (_message.Message,), {
  'DESCRIPTOR' : _JOBINFO,
  '__module__' : 'nbox_ws_pb2'
  # @@protoc_insertion_point(class_scope:jobs.ws.JobInfo)
  })
_sym_db.RegisterMessage(JobInfo)

ListJobsRequest = _reflection.GeneratedProtocolMessageType('ListJobsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTJOBSREQUEST,
  '__module__' : 'nbox_ws_pb2'
  # @@protoc_insertion_point(class_scope:jobs.ws.ListJobsRequest)
  })
_sym_db.RegisterMessage(ListJobsRequest)

ListJobsResponse = _reflection.GeneratedProtocolMessageType('ListJobsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTJOBSRESPONSE,
  '__module__' : 'nbox_ws_pb2'
  # @@protoc_insertion_point(class_scope:jobs.ws.ListJobsResponse)
  })
_sym_db.RegisterMessage(ListJobsResponse)

UploadCodeRequest = _reflection.GeneratedProtocolMessageType('UploadCodeRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADCODEREQUEST,
  '__module__' : 'nbox_ws_pb2'
  # @@protoc_insertion_point(class_scope:jobs.ws.UploadCodeRequest)
  })
_sym_db.RegisterMessage(UploadCodeRequest)

CreateJobRequest = _reflection.GeneratedProtocolMessageType('CreateJobRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEJOBREQUEST,
  '__module__' : 'nbox_ws_pb2'
  # @@protoc_insertion_point(class_scope:jobs.ws.CreateJobRequest)
  })
_sym_db.RegisterMessage(CreateJobRequest)

UpdateJobRequest = _reflection.GeneratedProtocolMessageType('UpdateJobRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEJOBREQUEST,
  '__module__' : 'nbox_ws_pb2'
  # @@protoc_insertion_point(class_scope:jobs.ws.UpdateJobRequest)
  })
_sym_db.RegisterMessage(UpdateJobRequest)

JobLogsRequest = _reflection.GeneratedProtocolMessageType('JobLogsRequest', (_message.Message,), {
  'DESCRIPTOR' : _JOBLOGSREQUEST,
  '__module__' : 'nbox_ws_pb2'
  # @@protoc_insertion_point(class_scope:jobs.ws.JobLogsRequest)
  })
_sym_db.RegisterMessage(JobLogsRequest)

JobLog = _reflection.GeneratedProtocolMessageType('JobLog', (_message.Message,), {
  'DESCRIPTOR' : _JOBLOG,
  '__module__' : 'nbox_ws_pb2'
  # @@protoc_insertion_point(class_scope:jobs.ws.JobLog)
  })
_sym_db.RegisterMessage(JobLog)

UpdateRunRequest = _reflection.GeneratedProtocolMessageType('UpdateRunRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATERUNREQUEST,
  '__module__' : 'nbox_ws_pb2'
  # @@protoc_insertion_point(class_scope:jobs.ws.UpdateRunRequest)
  })
_sym_db.RegisterMessage(UpdateRunRequest)

_WSJOBSERVICE = DESCRIPTOR.services_by_name['WSJobService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z0github.com/NimbleBoxAI/jobs-architecture/jobs_pb'
  _JOBINFO.fields_by_name['id']._options = None
  _JOBINFO.fields_by_name['id']._serialized_options = b'\030\001'
  _JOBINFO._serialized_start=165
  _JOBINFO._serialized_end=214
  _LISTJOBSREQUEST._serialized_start=216
  _LISTJOBSREQUEST._serialized_end=310
  _LISTJOBSRESPONSE._serialized_start=312
  _LISTJOBSRESPONSE._serialized_end=380
  _UPLOADCODEREQUEST._serialized_start=382
  _UPLOADCODEREQUEST._serialized_end=458
  _CREATEJOBREQUEST._serialized_start=460
  _CREATEJOBREQUEST._serialized_end=502
  _UPDATEJOBREQUEST._serialized_start=504
  _UPDATEJOBREQUEST._serialized_end=595
  _JOBLOGSREQUEST._serialized_start=597
  _JOBLOGSREQUEST._serialized_end=692
  _JOBLOG._serialized_start=694
  _JOBLOG._serialized_end=779
  _UPDATERUNREQUEST._serialized_start=781
  _UPDATERUNREQUEST._serialized_end=903
  _WSJOBSERVICE._serialized_start=906
  _WSJOBSERVICE._serialized_end=1502
# @@protoc_insertion_point(module_scope)



import os
import inspect
from enum import Enum
from threading import Thread, Lock
from time import sleep

import nbox.utils as U
from nbox.utils import logger

class OperatorType(Enum):
  """This Enum does not concern the user, however I am describing it so people can get a feel of the breadth
  of what nbox can do. The purpose of ``Operator`` is to build an abstract representation of any possible compute
  and execute them in any fashion needed to improve the overall performance of any distributed software system.
  Here are the different types of operators:

  #. ``UNSET``: this is the default mode and is like using vanilla python without any nbox features.
  #. ``JOB``: In this case the process is run as a batch process and the I/O of values is done using Relics
  #. ``SERVING``: In this case the process is run as an API proces
  #. ``WRAP_FN``: When we wrap a function as an ``Operator``, by default deployed as a job
  #. ``WRAP_CLS``: When we wrap a class as an ``Operator``, by default deployed as a serving
  """
  UNSET = "unset" # default
  JOB = "job"
  SERVING = "serving"
  WRAP_FN = "fn_wrap"
  WRAP_CLS = "cls_wrap"

  def _valid_deployment_types():
    return (OperatorType.JOB.value, OperatorType.SERVING.value)


class _UnsetSpec:
  def __init__(self):
    self.type = OperatorType.UNSET.value

  def __repr__(self):
    return f"[{self.type}]"

class _JobSpec:
  def __init__(
    self,
    job_id,
    rpc_fn_name,
    job,
    workspace_id,
  ):
    from nbox import Job

    self.type = OperatorType.JOB.value
    self.job_id = job_id
    self.rpc_fn_name = rpc_fn_name
    self.job: Job = job
    self.workspace_id = workspace_id

  def __repr__(self):
    return f"[{self.type} {self.job_id} {self.workspace_id}]"

class _ServingSpec:
  def __init__(
    self,
    serving_id,
    rpc_fn_name,
    fn_spec,
    workspace_id,
    track_io: bool = False
  ):
    self.type = OperatorType.SERVING.value
    self.serving_id = serving_id
    self.rpc_fn_name = rpc_fn_name
    self.fn_spec = fn_spec
    self.workspace_id = workspace_id
    self.track_io = track_io

  def __repr__(self):
    return f"[{self.type} {self.serving_id} {self.workspace_id}]"

class _WrapFnSpec:
  def __init__(self, fn_name, wrap_obj):
    self.type = OperatorType.WRAP_FN.value
    self.fn_name = fn_name
    self.wrap_obj = wrap_obj

  def ___repr__(self):
    return f"[{self.type} {self.fn_name}]"

class _WrapClsSpec:
  def __init__(self, cls_name, wrap_obj, init_ak):
    self.type = OperatorType.WRAP_CLS.value
    self.cls_name = cls_name
    self.wrap_obj = wrap_obj
    self.init_ak = init_ak

  def ___repr__(self):
    return f"[{self.type} {self.cls_name}]"


def get_operator_location(_op):
  # get the filepath and name to import for convience
  if _op._op_type == OperatorType.UNSET:
    fp = inspect.getfile(_op.__class__)
    name = _op.__class__.__qualname__
  elif _op._op_type in [OperatorType.JOB, OperatorType.SERVING]:
    raise ValueError("Cannot deploy an operator that is already deployed")
  elif _op._op_type == OperatorType.WRAP_FN:
    fp = _op.__file__
    name = _op.__qualname__[3:] # to account for "fn_"
  elif _op._op_type == OperatorType.WRAP_CLS:
    fp = _op.__file__
    name = _op.__qualname__[4:] # to account for "cls_"
  fp = os.path.abspath(fp) # get the abspath, will be super useful later
  folder, file = os.path.split(fp)
  return fp, folder, file, name


class _LocalMapFuture:
  def __init__(self, result_folder, tag):
    self.fp = U.join(result_folder, tag + "_output")

  def __repr__(self) -> str:
    return f"_LocalMapFuture ({self.fp})"

  def wait(self):
    while not self.done():
      sleep(0.1)

  def done(self):
    os.path.exists(self.fp)

  def result(self):
    if not self.done():
      logger.error("Cannot get result before it is done")
      return None
    return U.from_pickle(self.fp)


class _LocalMapPooler:
  def __init__(self, folder, file, name, n_workers = 10):
    self.folder = folder
    self.file = file
    self.name = name
    self.n_workers = n_workers
    
    # things to manage the processes
    self.items = []
    self.processes = []
    self.running = {}
    self.completed = 0
    self._mutex = Lock()
    self._break = False

    self.event_loop = Thread(target = self._event_loop, daemon = True)

  def __repr__(self):
    return f"_LocalMapPooler ({self.n_workers:03d} workers) running '{self.folder}/{self.file}::{self.name}'"

  def _there_is_space(self):
    return len(self.processes) < self.n_workers

  def _event_loop(self):
    # this sees what is in the queue if any slot is finished and runs it
    while True:
      with self._mutex:
        n_items = len(self.items)
        _break = self._break
      print(_break, n_items)
      if _break:
        return
      if not n_items:
        sleep(0.1)

      status = [p.poll() for p in self.processes]
      success = [x == 0 for x in status]
      errored = [x is not None and x != 0 for x in status]
      completed = sum([s is not None for s in status])
      logger.debug(f"done: [{completed}/{len(status)}], success: {sum(success)}, errored: {sum(errored)}")

      pending = n_items - completed
      while pending and self._there_is_space():
        with self._mutex:
          try:
            (tag, fn) = self.items.pop(0)
          except IndexError:
            break

        print("Processing tag", tag)
        p = fn()
        self.processes.append(p)
        pending -= 1

  def submit(self, tag, fn, result_folder) -> _LocalMapFuture:
    self.items.append((tag, fn))
    fut = _LocalMapFuture(result_folder, tag)
    return fut

  def start(self):
    self.event_loop.start()

  def stop(self):
    with self._mutex:
      self._break = True
    self.event_loop.join()



"""
With NimbleBox you can run cluster wide workloads from anywhere. This requires capabilities around distributed computing,
process managements. The code here is tested along with ``nbox.Relic`` to perform distributed local processing.
"""

import os

import nbox.utils as U
from nbox import RelicsNBX
from nbox.auth import secret, ConfigString
from nbox import Operator, logger
from nbox.utils import SimplerTimes
from nbox.nbxlib.tracer import Tracer
from nbox.hyperloop.job_pb2 import Job
from nbox import Operator, nbox_grpc_stub
from nbox.messages import rpc
from nbox.hyperloop.nbox_ws_pb2 import UpdateRunRequest
from nbox.nbxlib.serving import serve_operator


# Manager
class LocalNBXLet(Operator):
  def __init__(self, op: Operator, in_key: str, out_key: str):
    super().__init__()
    self.op = op
    self.in_key = in_key
    self.out_key = out_key

  # def start_status(self):
  #   def worker():
  #     # function to keep on updating the cache and seeing the progress
  #     while True:
  #       available = 0
  #       for key in self.keys:
  #         if self.relic.has(key):
  #           available += 1
  #       if self.verbose:
  #         logger.info(f"Available: [{available}/{len(self.keys)}]")
  #       sleep(0.69)

  #   # start the status thread
  #   self.cache_man = Thread(target = worker, daemon=True)
  #   self.cache_man.start()

  # def prepare_children(self):
  #   self.op_agents = {}
  #   self.keys = set()
  #   for op_name, res in self.op._op_to_resource_map.items():
  #     if res != None:
  #       # replace the Opeartors in the graph with AgentsStubs
  #       op = getattr(self.op, op_name)
  #       agent_op = AgentOpRootStub(op, op_name, res, self.relic)
  #       setattr(self.op, op_name, agent_op)
  #       self.op_agents[op_name] = agent_op
  #       self.keys.add(op_name + "_input")
  #       self.keys.add(op_name + "_output")

  def __repr__(self):
    return f"LocalNBXLet({self.op.__qualname__}, {self.in_key}, {self.out_key})"

  def forward(self):
    x = U.from_pickle(self.in_key)
    y = self.op(*x)
    U.to_pickle(y, self.out_key)


class NBXLet(Operator):
  def __init__(self, op: Operator):
    """The Operator that runs the things on any pod on the NimbleBox Jobs + Deploy platform.
    Name mimics the kubelet, dockerlet, raylet, etc"""
    super().__init__()
    self.op = op

  def run(self):
    """Run this as a batch process"""
    tracer = Tracer()

    self.op.propagate(_tracer = tracer)
    if hasattr(self.op._tracer, "job_proto"):
      self.op.thaw(self.op._tracer.job_proto)

    secret.put("username", tracer.job_proto.auth_info.username)
    workspace_id = tracer.job_proto.auth_info.workspace_id
    secret.put(ConfigString.workspace_id.value, workspace_id, True)
    job_id = tracer.job_id
    status = Job.Status.ERROR

    try:
      # now for some jobs there might be a relic Object so we can check if that exists, it will always
      # be present in the dot_deploy_cache folder and will be in the {job_id} folder
      relic = RelicsNBX("cache", workspace_id, create = True)

      # check if there is a specific relic for this job
      run_tag = os.getenv("NBOX_RUN_METADATA", "")
      _in = f"{job_id}/args_kwargs"
      if run_tag:
        _in += f"_{run_tag}"
      if relic.has(_in):
        (args, kwargs) = relic.get_object(_in)
      else:
        args, kwargs = (), {}
      out = self.op(*args, **kwargs)
      _out = f"{job_id}/return"
      if run_tag:
        _out += f"_{run_tag}"
      relic.put_object(_out, out)
      status = Job.Status.COMPLETED
    except Exception as e:
      U.log_traceback()
    finally:
      logger.info(f"Job {job_id} completed with status {status}")
      if hasattr(tracer, "job_proto"):
        self.op._tracer.job_proto.status = status
        rpc(
          nbox_grpc_stub.UpdateRun, UpdateRunRequest(
            token = tracer.run_id, job = tracer.job_proto, updated_at = SimplerTimes.get_now_pb()
          ), "Failed to end job!"
        )
      U._exit_program()

  def serve(self, host: str = "0.0.0.0", port: int = 8000, *, model_name: str = None):
    """Run a serving API endpoint"""
    try:
      serve_operator(self.op, host = host, port = port, model_name = model_name)
    except Exception as e:
      U.log_traceback()
      logger.error(f"Failed to serve operator: {e}")
      U._exit_program()
